filename = "line.msh"
h = 0.1
hf = 0.01

import gmsh
gmsh.initialize()
gmsh.option.setNumber("General.Verbosity", 0)
gmsh.option.setNumber("Mesh.MshFileVersion", 2.2)

gmsh.model.add(filename)

geo = gmsh.model.geo
p1 = geo.addPoint(0,  0, 0, h)
p2 = geo.addPoint(1,  0, 0, h)
p3 = geo.addPoint(1, .5, 0, hf)
p4 = geo.addPoint(1,  1, 0, h)
p5 = geo.addPoint(0,  1, 0, h)
p6 = geo.addPoint(0, .5, 0, hf)

l1 = geo.addLine(p1, p2)
l2 = geo.addLine(p2, p3)
l3 = geo.addLine(p3, p4)
l4 = geo.addLine(p4, p5)
l5 = geo.addLine(p5, p6)
l6 = geo.addLine(p6, p1)

lf = geo.addLine(p3, p6)

cl = geo.addCurveLoop([l1, l2, l3, l4, l5, l6])
s1 = geo.addPlaneSurface([cl])

geo.synchronize()
gmsh.model.mesh.embed(1, [lf], 2, s1)

gmsh.model.mesh.generate(dim=2)
gmsh.write(filename)
gmsh.finalize()



import logging, traceback
logger = logging.getLogger(__name__)

from dune.fem import adapt

def moveMesh(grid, gridFunctions, getShifts=None, igrid=None, igridFunctions=[]):
    """Move the mesh and adapt given discrete functions.

    Args:
        grid: The adaptive bulk grid view.
        gridFunctions: List of bulk discrete functions to be adapted.
        getShifts: Callback function to obtain a list with shift per interface vertex index.
        igrid (optional): The adaptive interface grid view.
        igridFunctions (optional): List of interface discrete functions to be adapted.

    Returns:
        Number of ensure and mark steps.
    """
    hgrid = grid.hierarchicalGrid

    ensure = 0
    if getShifts is not None:
        shifts = getShifts()

        while hgrid.ensureInterfaceMovement(2 * shifts) and ensure < 10:
            ensure += 1
            if len(gridFunctions) > 0:
                adapt(gridFunctions)
            if len(igridFunctions) > 0:
                adapt(igridFunctions)
            shifts = getShifts()

        hgrid.moveInterface(shifts)

    mark = 0
    while hgrid.markElements() and mark < 10:
        mark += 1
        if len(gridFunctions) > 0:
            adapt(gridFunctions)
        if len(igridFunctions) > 0:
            adapt(igridFunctions)

    return {"ensure": ensure, "mark": mark}



from dune.grid import reader
from dune.mmesh import mmesh
from dune.fem import adapt
from dune.fem.view import adaptiveLeafGridView as adaptive
from dune.fem.space import dglagrange
from dune.fem.function import integrate

from ufl import SpatialCoordinate, conditional

from dune.mmesh.test.grids import line
grid = mmesh((reader.gmsh, line.filename), 2)
hgrid = grid.hierarchicalGrid
gridView = adaptive(grid)
igridView = adaptive(hgrid.interfaceGrid)

space = dglagrange(gridView, order=1)
x = SpatialCoordinate(space)
uh = space.interpolate(x[0]+x[1], name="uh")

ispace = dglagrange(igridView, order=1)
iuh = ispace.interpolate(x[0], name="iuh")

def addSomeInterface(number):
  for e in gridView.elements:
    for i in gridView.intersections(e):
      if hgrid.isInterface(i) or i.boundary:
        continue
      if number == 0:
        print("Add interface at", i.geometry.center)
        hgrid.addInterface(i)
        return
      number -= 1

addSomeInterface(0)
addSomeInterface(10)
addSomeInterface(20)
addSomeInterface(42)
adapt([uh])
adapt([iuh])

# Initialize new elements with 0
iuh.interpolate(conditional(abs(iuh - x[0]) < 1e-6, iuh, 0))

gridView.writeVTK("adaptation", pointdata=[uh], nonconforming=True)
igridView.writeVTK("adaptation-interface", pointdata=[iuh], nonconforming=True)

intuh = integrate(gridView, uh, order=1)
intiuh = integrate(igridView, iuh, order=1)
print(intuh, intiuh)
assert(abs(intuh - 1) < 1e-6)
assert(abs(intiuh - 0.5) < 1e-6)



from ._grids import *
from ._move import *
from ._skeletontrace import *
from ._solve import *
from .utility import *

registry = dict()
registry["grid"] = grid_registry

from dune.fem import parameter
parameter.append({"fem.adaptation.method": "callback"})



# Generate mesh
import gmsh
gmsh.initialize()
gmsh.option.setNumber("Mesh.MshFileVersion", 2.2)
name = "grid.msh"
h = 0.01
gmsh.model.add(name)
kernel = gmsh.model.occ
box = kernel.addRectangle(0, 0, 0, 1, 1)
p0 = kernel.addPoint(0.5, 0.5, 0, h)
p1 = kernel.addPoint(0.25, 0.5, 0, h)
p2 = kernel.addPoint(0.75, 0.75, 0, h)
p3 = kernel.addPoint(0.75, 0.25, 0, h)
lf1 = kernel.addLine(p0, p1)
lf2 = kernel.addLine(p0, p2)
lf3 = kernel.addLine(p0, p3)
kernel.synchronize()
gmsh.model.mesh.embed(1, [lf1, lf2, lf3], 2, box)
gmsh.model.mesh.generate(dim=2)
gmsh.write(name)
gmsh.finalize()


# Grid creation
from dune.grid import reader
from dune.mmesh import mmesh
gridView  = mmesh((reader.gmsh, name), 2)
igridView = gridView.hierarchicalGrid.interfaceGrid


# Coupled problem
from dune.mmesh import trace, skeleton, interfaceIndicator, monolithicSolve

from dune.fem.space import dglagrange, lagrange
space = dglagrange(gridView, order=3)
ispace = lagrange(igridView, order=3)

from ufl import *
u = TrialFunction(space)
v = TestFunction(space)
n = FacetNormal(space)
h = MinFacetEdgeLength(space)
uh = space.interpolate(0, name="uh")

iu = TrialFunction(ispace)
iv = TestFunction(ispace)
iuh = ispace.interpolate(0, name="iuh")

from dune.mmesh import interfaceIndicator
I = interfaceIndicator(igridView)

from dune.ufl import Constant
q = Constant(1, name="q")
beta = Constant(1e2, name="beta")
omega = Constant(1e-6, name="omega")

a  = inner(grad(u), grad(v)) * dx
a += beta / h * inner(jump(u), jump(v)) * (1-I)*dS
a -= dot(dot(avg(grad(u)), n('+')), jump(v)) * (1-I)*dS

a += beta / h * inner(u - 0, v) * ds
a -= dot(dot(grad(u), n), v) * ds

ia  = inner(grad(iu), grad(iv)) * dx
ib  = q * iv * dx

from dune.mmesh import skeleton, trace
a -= (skeleton(iuh)('+') - u('+')) / omega * v('+') * I*dS
a -= (skeleton(iuh)('-') - u('-')) / omega * v('-') * I*dS

ia += (iu - trace(uh)('+')) / omega * iv * dx
ia += (iu - trace(uh)('-')) / omega * iv * dx

from dune.fem.scheme import galerkin
scheme = galerkin([a == 0])
ischeme = galerkin([ia == ib])

from dune.mmesh import monolithicSolve
monolithicSolve(schemes=(scheme, ischeme), targets=(uh, iuh), verbose=True)

# Plot
gridView.writeVTK("example", pointdata=[uh], nonconforming=True)
igridView.writeVTK("interface", pointdata=[iuh])

import matplotlib.pyplot as plt
from dune.fem.plotting import plotPointData as plot
fig = plt.figure()
plot(uh, linewidth=0, clim=[0, 0.17], figure=fig)
plot(iuh, linewidth=0.01, colorbar=None, clim=[0, 0.17], figure=fig)
plt.savefig("plot.png")



# A rectangle grid with a vertical interface

name = "vertical.msh"
h = 0.05
hf = 0.05

import gmsh
gmsh.initialize()
gmsh.option.setNumber("General.Verbosity", 0)
gmsh.option.setNumber("Mesh.MshFileVersion", 2.2)

gmsh.model.add(name)

p1 = gmsh.model.geo.addPoint(   0, 0, 0, h, 1)
p2 = gmsh.model.geo.addPoint( 0.5, 0, 0, hf, 2)
p3 = gmsh.model.geo.addPoint(   1, 0, 0, h, 3)
p4 = gmsh.model.geo.addPoint(   1, 1, 0, h, 4)
p5 = gmsh.model.geo.addPoint( 0.5, 1, 0, hf, 5)
p6 = gmsh.model.geo.addPoint(   0, 1, 0, h, 6)

l1 = gmsh.model.geo.addLine(p1, p2, 1)
l2 = gmsh.model.geo.addLine(p2, p3, 2)
l3 = gmsh.model.geo.addLine(p3, p4, 3)
l4 = gmsh.model.geo.addLine(p4, p5, 4)
l5 = gmsh.model.geo.addLine(p5, p6, 5)
l6 = gmsh.model.geo.addLine(p6, p1, 6)
lf = gmsh.model.geo.addLine(p2, p5, 10)

gmsh.model.geo.addCurveLoop([1, 10, 5, 6], 1)
gmsh.model.geo.addPlaneSurface([1], 0)
gmsh.model.geo.addCurveLoop([2, 3, 4, -10], 2)
gmsh.model.geo.addPlaneSurface([2], 1)

gmsh.model.geo.synchronize()

gmsh.model.mesh.generate(dim=2)
gmsh.write(name)
gmsh.finalize()



# Grid creation
from dune.grid import reader
from dune.mmesh import mmesh
dim = 2
file = "horizontal.msh"
gridView  = mmesh((reader.gmsh, file), dim)
igridView = gridView.hierarchicalGrid.interfaceGrid



# Bulk problem
from ufl import *
from dune.ufl import DirichletBC
from dune.fem.space import lagrange
from dune.fem.scheme import galerkin
from dune.fem.function import integrate
import time

space = lagrange(gridView, order=3)
u = TrialFunction(space)
v = TestFunction(space)

x = SpatialCoordinate(space)
exact = sin(x[0]*x[1]*4*pi)
f = -div(grad(exact))

a = inner(grad(u), grad(v)) * dx
b = f * v * dx

start_time = time.time()
scheme = galerkin([a == b, DirichletBC(space, exact)], solver=("suitesparse", "umfpack"))
print("bulk:", time.time() - start_time)
uh = space.interpolate(0, name="solution")
scheme.solve(target=uh)

def L2(u1, u2):
    return sqrt(integrate(u1.grid, (u1-u2)**2, order=5))

assert( L2(uh, exact) < 1e-6 )



# Interface problem
ispace = lagrange(igridView, order=3)
iuh = ispace.interpolate(0, name="isolution")

iu = TrialFunction(ispace)
iv = TestFunction(ispace)

ix = SpatialCoordinate(ispace)
iexact = sin(0.5*ix[dim-2]*4*pi)
iF = -div(grad(iexact))

ia = inner(grad(iu), grad(iv)) * dx
ib = iF * iv * dx

start_time = time.time()
ischeme = galerkin([ia == ib, DirichletBC(ispace, iexact)])
print("interface:", time.time() - start_time)
ischeme.solve(target=iuh)
assert( L2(iuh, iexact) < 1e-6 )



# Couple bulk to surface
from dune.mmesh import trace
tr = avg(trace(uh))
ib = inner(grad(tr), grad(iv)) * dx

iuh.interpolate(0)
ischeme = galerkin([ia == ib, DirichletBC(ispace, avg(trace(uh)))])
ischeme.solve(target=iuh)
assert( L2(iuh, iexact) < 1e-6 )



# Couple surface to bulk
from dune.mmesh import skeleton
sk = skeleton(iuh)
b = avg(sk) * avg(v) * dS

uh.interpolate(0)
scheme = galerkin([a == b, DirichletBC(space, 0)])
scheme.solve(target=uh)



# Coupled solve
from dune.mmesh import monolithicSolve, interfaceIndicator
from dune.fem.space import dglagrange
from dune.ufl import Constant

from dune.fem.view import geometryGridView
from dune.fem.function import uflFunction
x = SpatialCoordinate(triangle)

vectorSpace = dglagrange(gridView, dimRange=2, order=1)
position = vectorSpace.interpolate(x, name="position")
geoGrid = geometryGridView(position)

space = dglagrange(geoGrid, order=3)
u = TrialFunction(space)
v = TestFunction(space)

ispace = lagrange(igridView, order=3)
iu = TrialFunction(ispace)
iv = TestFunction(ispace)

uh = space.interpolate(0, name="solution")
iuh = ispace.interpolate(0, name="isolution")

I = interfaceIndicator(igridView, grid=geoGrid)
n = FacetNormal(space)
n_g = FacetNormal(ispace)
beta = Constant(1e2, name="beta")
omega = Constant(1e-6, name="omega")

a  = inner(grad(u), grad(v)) * dx
a += beta * inner(jump(u), jump(v)) * (1-I)*dS
a -= dot(dot(avg(grad(u)), n('+')), jump(v)) * (1-I)*dS

a += beta * inner(u - 0, v) * ds
a -= dot(dot(grad(u), n), v) * ds

sk = skeleton(iuh, grid=geoGrid)
a -= (sk('+') - u('+')) / omega * v('+') * I*dS
a -= (sk('-') - u('-')) / omega * v('-') * I*dS


ia  = inner(grad(iu), grad(iv)) * dx
ia += beta * inner(jump(iu), jump(iv)) * dS
ia -= inner(avg(grad(iu)), n_g('+')) * jump(iv) * dS

tr  = trace(uh)
ia += (iu - tr('+')) / omega * iv * dx
ia += (iu - tr('-')) / omega * iv * dx

ib  = 1 * iv * dx

scheme = galerkin([a == 0])
ischeme = galerkin([ia == ib])

start_time = time.time()
monolithicSolve(schemes=(scheme, ischeme), targets=(uh, iuh), verbose=True)
print("monolithicSolve:", time.time() - start_time)
assert( (max(iuh.as_numpy) - 0.14063) < 1e-3 )



import sys
import os
import glob
import subprocess

if __name__ == '__main__':
    print("===  Testing Dune-MMesh  ===")

    path = os.path.dirname(os.path.realpath(__file__))
    tests = glob.glob(path + '/[!_]*.py')

    exclude = ["boundary.py"]
    tests = [t for t in tests if not t.endswith(tuple(exclude))]

    N = len(tests)
    passed = 0
    i = 0
    for test in tests:
        i += 1
        script = os.path.split(test)[1].split('.')[0]
        print("Test "+str(i)+"/"+str(N)+":", script)
        try:
            subprocess.run([sys.executable, test], check=True)
            print(" - passed")
            passed += 1
        except subprocess.CalledProcessError:
            print(" - failed!")

    print("=== ", str(passed)+"/"+str(N), "tests passed ===")



#!/usr/bin/env python
# coding: utf-8

# # Poroelasticity

# We consider quasi-static linear Biot-equations with an interface $\Gamma$ considered as thin heterogeneity with different permeability.
#
# Find $\mathbf{u}, p, p_\Gamma$ s.t
# \begin{align}
# - \operatorname{div}( \mathbf{K} \nabla p ) &= 0, \quad \text{in } \Omega, \\
# - \operatorname{div} \left( \sigma( \mathbf{u} ) - \alpha p I \right) &= 0, \quad \text{in } \Omega, \\
# - \operatorname{div}( \mathbf{K}_\Gamma \nabla p_\Gamma ) &= q_\Gamma, \quad \text{in } \Gamma, \\
# \end{align}
# where
# \begin{align}
# \sigma( \mathbf{u} ) &= \lambda \operatorname{tr} \epsilon ( \mathbf{u} ) I + 2 \mu \epsilon( \mathbf{u} ), \\
# \epsilon( \mathbf{u} ) &= \frac{1}{2} \left( \nabla \mathbf{u} + (\nabla \mathbf{u})^T \right) \\
# \end{align}
#
# and $\lambda, \mu > 0$ are the Lamé constants, $\alpha$ the Biot-Willis constant, $\mathbf{K}$, $\mathbf{K}_\Gamma$ bulk and interface permeability and $q_\Gamma$ a source term.
#
# We impose interior boundary conditions
# \begin{align}
# p &= p_\Gamma, \\
# \left( \sigma(\mathbf{u}) - \alpha p I \right) \cdot \mathbf{n} &= -p_\Gamma \mathbf{n}.
# \end{align}
# and consider a [T-junction](grids/tjunction.rst) as grid geometry.

# In[1]:


from dune.grid import reader
from dune.mmesh import mmesh

file = "grids/tjunction.msh"

gridView = mmesh((reader.gmsh, file), 2)
igridView = gridView.hierarchicalGrid.interfaceGrid

# In[2]:


from ufl import *
from dune.ufl import Constant

lamb    = Constant( 1.2, name="lambda")
mu      = Constant( 0.8, name="mu")
alpha   = Constant(   1, name="alpha")
K       = as_matrix([[1e-4, 0], [0, 1e-6]])
K_g     = Constant(   1, name="K_g")
q       = Constant(1e-6, name="q")

epsilon = lambda u: 0.5 * (nabla_grad(u) + nabla_grad(u).T)
sigma   = lambda u: lamb * div(u) * Identity(2) + 2 * mu * epsilon(u)

# In[3]:


from dune.fem.space import dglagrange

# Bulk
space = dglagrange(gridView, dimRange=3, order=1)

trial = TrialFunction(space)
test = TestFunction(space)

p,  ux,  uy  = split(trial)
pp, uux, uuy = split(test)

u  = as_vector([ux,  uy ])
uu = as_vector([uux, uuy])

x = SpatialCoordinate(space)
n = FacetNormal(space)
h = FacetArea(space)

left = conditional(x[0] < 1e-6, 1, 0)
right = conditional(x[0] > 1-1e-6, 1, 0)
bottom = conditional(x[1] < 1e-6, 1, 0)
top = conditional(x[1] > 1-1e-6, 1, 0)

solution = space.interpolate([0,0,0], name="solution")

# Interface
space_g = dglagrange(igridView, order=1)

p_g  = TrialFunction(space_g)
pp_g = TestFunction(space_g)

x_g = SpatialCoordinate(space_g)
n_g = FacetNormal(space_g)
h_g = FacetArea(space_g)

solution_g = space_g.interpolate(0, name="solution_g")

# ## Interior Penalty Discontinuous Galerkin (IPDG) Scheme
#
# We implement a mixed-dimensional Interior Penalty Discontinuous Galerkin (IPDG) scheme which directly allows for a discontinuity along the interface. Including the penalty and consistency terms, we have the following weak form.
#
# \begin{align*}
# \renewcommand{\jump}[1]{[\mskip-5mu[ #1 ]\mskip-5mu]}
# \renewcommand{\avg}[1]{\{\mskip-5mu\{ #1 \}\mskip-5mu\}}
# \end{align*}
#
# \begin{align}
# A(\mathbf{u}, p; \mathbf{v}, \varphi) :=& \int_\Omega ( \mathbf{K} \nabla p ) \cdot \nabla \varphi~dx
# - \int_\Omega q \varphi~dx \\
# &+ \int_{\mathcal{F} \setminus \mathcal{F}_\Gamma} \frac{\beta}{h} \jump{p} \jump{\varphi}
# - \avg{ \mathbf{K} \nabla p \cdot \mathbf{n} } \jump{\varphi}~dS
# + \int_{\mathcal{F}_D^p} \frac{\beta}{h} (p-p_D) \varphi - \mathbf{K} \nabla p \cdot \mathbf{n}\varphi~ds \\
# &+ \int_{\mathcal{F}_\Gamma} \frac{\beta}{h} (p^+ - p_\Gamma) \varphi^+ - \mathbf{K} \nabla p^+ \cdot \mathbf{n}^+\varphi^+~dS
# + \int_{\mathcal{F}_\Gamma} \frac{\beta}{h} (p^- - p_\Gamma) \varphi^- - \mathbf{K} \nabla p^- \cdot \mathbf{n}^-\varphi^-~dS \\
# %
# &+ \int_\Omega (\sigma(\mathbf{u}) - \alpha p I) : \epsilon(\mathbf{v})~dx
# + \int_{\mathcal{F} \setminus \mathcal{F}_\Gamma} \frac{\beta}{h} \jump{\mathbf{u}} \jump{\mathbf{v}} - ( \avg{ \sigma(\mathbf{u}) - \alpha p I } \cdot \mathbf{n}) \cdot \jump{\mathbf{v}}~dS \\
# & + \int_{\mathcal{F}_D^u} \frac{\beta}{h} (\mathbf{u} - \mathbf{u}_D )\cdot \mathbf{v} - ( ( \sigma(\mathbf{u}) - \alpha p I ) \cdot \mathbf{n}) \cdot \mathbf{v}~ds \\
# & - \int_{\mathcal{F}_\Gamma} - p_\Gamma (\mathbf{v}^+ \cdot \mathbf{n}^+) - p_\Gamma (\mathbf{v}^- \cdot \mathbf{n}^-)~dS \\
# %
# A_\Gamma(p_\Gamma; \varphi_\Gamma) :=&
# \int_\Gamma (\mathbf{K}_\Gamma \nabla p_\Gamma) \cdot \nabla \varphi_\Gamma - q \varphi_\Gamma + \beta (p_\Gamma - \avg{p}) \varphi_\Gamma~dx \\
# &+ \int_{\mathcal{F}^\Gamma} \frac{\beta}{h_\Gamma} \jump{p_\Gamma} \jump{\varphi_\Gamma}
# - \avg{ \mathbf{K}_\Gamma \nabla p_\Gamma \cdot \mathbf{n}_\Gamma } \jump{\varphi_\Gamma}~dS
# \end{align}
#
# This weak form is implemented using UFL and Dune-MMesh's trace and skeleton functionality as follows.

# In[4]:


from dune.mmesh import skeleton, trace, interfaceIndicator

I = interfaceIndicator(igridView)
beta = Constant(1e2, name="beta")

p_gamma = avg(skeleton(solution_g))
tracep = trace(solution)[0]


# Pressure
a  = inner(K * grad(p), grad(pp)) * dx
a -= q * pp * dx

a += beta / h * inner(jump(p), jump(pp)) * (1-I)*dS
a -= inner(avg(K * grad(p)), n('+')) * jump(pp) * (1-I)*dS

a += beta / h * (p - 0) * pp * (left+right) * ds
a -= inner(K * grad(p), n) * pp * (left+right) * ds


# Interface pressure
a_g  = inner(K_g * grad(p_g), grad(pp_g)) * dx
a_g -= q * pp_g * dx

a_g += beta / h_g * inner(jump(p_g), jump(pp_g)) * dS
a_g -= inner(avg(K_g * grad(p_g)), n_g('+')) * jump(pp_g) * dS


# Pressure is continuous at the interface
a += beta * (p('+') - p_gamma) * pp('+') * I*dS
a -= inner(K * grad(p('+')), n('+')) * pp('+') * I*dS

a += beta * (p('-') - p_gamma) * pp('-') * I*dS
a -= inner(K * grad(p('-')), n('-')) * pp('-') * I*dS

a_g += beta * (p_g - avg(tracep)) * pp_g * dx


# Displacement
sigma_p = lambda u, p: sigma(u) - alpha * p * Identity(2)
a += inner(sigma_p(u, p), epsilon(uu)) * dx

a += beta / h * inner(jump(u), jump(uu)) * (1-I)*dS
a -= dot(dot(avg(sigma_p(u, p)), n('+')), jump(uu)) * (1-I)*dS

a += beta / h * inner(u - as_vector([0,0]), uu) * (top+bottom) * ds
a -= dot(dot(sigma_p(u, p), n), uu) * (top+bottom) * ds


# Normal stress is -p at the interface
a -= -p_gamma * inner(uu('+'), n('+')) * I*dS
a -= -p_gamma * inner(uu('-'), n('-')) * I*dS

# ## Monolithic solution strategy
#
# We use the `monolithicSolve` to obtain the solution of the strongly coupled mixed-dimensional system.

# In[5]:


from dune.fem.scheme import galerkin
scheme   = galerkin([a   == 0], solver=("suitesparse", "umfpack"))
scheme_g = galerkin([a_g == 0], solver=("suitesparse", "umfpack"))

from dune.mmesh import monolithicSolve
converged = monolithicSolve(schemes=(scheme, scheme_g), targets=(solution, solution_g), verbose=True)

# In[6]:


import matplotlib.pyplot as plt
from dune.fem.plotting import plotPointData as plot

fig, axs = plt.subplots(1, 4, figsize=(20,5))
axs[0].set_title('Pressure')
plot(solution[0], figure=(fig, axs[0]), gridLines=None)
axs[1].set_title('Displacement (x)')
plot(solution[1], figure=(fig, axs[1]), gridLines=None)
axs[2].set_title('Displacement (y)')
plot(solution[2], figure=(fig, axs[2]), gridLines=None)
axs[3].set_title('Displacement')
plot(solution, vectors=[1,2], figure=(fig, axs[3]), gridLines=None)



from dune.grid import reader
from dune.mmesh import mmesh
from dune.mmesh import interfaceIndicator
from dune.fem.space import lagrange
from dune.fem.scheme import galerkin
from dune.fem.function import integrate
from dune.fem.view import geometryGridView
from ufl import *

# 2D
from dune.mmesh.test.grids import line
grid2 = mmesh((reader.gmsh, line.filename), 2)

# 3D
from dune.mmesh.test.grids import plane
grid3  = mmesh((reader.gmsh, plane.filename), 3)

# 2D GeometryGrid
space = lagrange(grid2, order=1, dimRange=2)
x = SpatialCoordinate(space)
disp = space.interpolate(x, name="disp")
geoGrid = geometryGridView(disp)


for grid in [grid2, grid3, geoGrid]:
  igrid = grid.hierarchicalGrid.interfaceGrid
  I = interfaceIndicator(igrid, grid=grid)

  space = lagrange(grid, order=1)
  u  = TrialFunction(space)
  uu = TestFunction(space)
  n = FacetNormal(space)
  x = SpatialCoordinate(space)
  topbottom = conditional(x[1] < 1e-6, 1, 0) + conditional(x[1] > 1-1e-6, 1, 0)

  A  = inner(grad(u), grad(uu)) * dx
  A += 1e6 * (avg(u) - 1) * avg(uu) * I*dS
  A += 1e6 * (u - 0) * uu * topbottom * ds

  scheme = galerkin([A == 0])
  uh = space.function(name="uh")
  scheme.solve(uh)

  grid.writeVTK("interfaceindicator", pointdata=[uh])

  intU = integrate(grid, uh, order=1)
  print(intU)
  assert(abs(intU - 0.5) < 1e-2)



#!/usr/bin/env python
# coding: utf-8

# # Two-phase Navier-Stokes with surface tension

# Let's consider the incompressible Navier-Stokes equations with two immiscible phases and surface tension to demonstrate more complex usage of Dune-MMesh's capabilities.
#
# A domain $\Omega \subset \mathbb{R}^2$ is assumed to be separeted into two phases $\Omega_i(t), i=1,2,$ by a sharp interface $\Gamma(t)$.
#
# Find $(u, p)$ and $\Gamma(t)$ s.t.
# \begin{align}
# \renewcommand{\jump}[1]{[\mskip-5mu[ #1 ]\mskip-5mu]}
# \end{align}
# \begin{align}
# \rho u_t + \nabla \cdot (\rho u \otimes u) + \nabla \cdot T(u, p) &= 0, &\qquad &\text{in } \Omega_i(t), \quad &i&=1,2,\\
# \nabla \cdot u &= 0, &\qquad &\text{in } \Omega_i(t), \quad &i&=1,2,\\
# \jump{ p } &= \sigma \kappa \cdot n, &\qquad &\text{on } \Gamma(t),\\
# \jump{u} &= 0, &\qquad &\text{on } \Gamma(t),\\
# \dot x &= u, &\qquad &x \in \Gamma(t),\\
# u(0) &= u_0, &\qquad &\text{in } \Omega_i(0), \quad &i&=1,2,\\
# \Gamma(0) &= \Gamma_0.
# \end{align}
#
# Here, $T(u, p) := p I - \mu (\nabla u + (\nabla u)^T))$ is the stress tensor, $\mu_i, i=1,2,$ are the dynamic viscositities and $\rho_i, i=1,2,$ the densities of the two phases, $\sigma$ is the surface tension and $\kappa$ is the signed mean curvature of the interface times its normal.
#
# Now, we use Dune-MMesh to compute the dynamics of a droplet in a channel. Let us consider the [circle](grids/circle.rst) grid.

# In[1]:


from dune.grid import reader
from dune.mmesh import mmesh, trace, skeleton, domainMarker
from dune.fem.view import adaptiveLeafGridView as adaptive

dim = 2
file = "grids/circle.msh"

gridView = adaptive( mmesh((reader.gmsh, file), dim) )
hgrid = gridView.hierarchicalGrid
igridView = adaptive( hgrid.interfaceGrid )

# In[2]:


from ufl import *
from dune.ufl import Constant

mu0   = Constant(   1, name="mu0")
mu1   = Constant(   1, name="mu1")
rho0  = Constant(   1, name="rho0")
rho1  = Constant(   2, name="rho1")
sigma = Constant(0.03, name="sigma")

dt = 0.05
T  = 12.0
tau  = Constant(dt, name="tau")

# ## Domain markers
#
# The domain markers are initialized by the physical identifiers passed from the mesh file. We use them to interpolate the phase’s parameters.

# In[3]:


from dune.mmesh import domainMarker

dm = domainMarker(gridView)
mu  = (1-dm) * mu0  + dm * mu1
rho = (1-dm) * rho0 + dm * rho1

# In[4]:


from dune.fem.space import lagrange, dglagrange

pspace = dglagrange(gridView, order=1)
p = TrialFunction(pspace)
q = TestFunction(pspace)

uspace = dglagrange(gridView, dimRange=dim, order=2)
u = TrialFunction(uspace)
v = TestFunction(uspace)

ph = pspace.interpolate(0, name="ph")
uh = uspace.interpolate([0,0], name="uh")
uh1 = uspace.interpolate([0,0], name="uh1")

# ## Curvature
#
# The mean curvature $\kappa$ of the interface times its normal can be computed by solving
# \begin{align*}
# \int_{\Gamma} \kappa \cdot \phi + \nabla x \cdot \nabla \phi \ dS = 0, &\qquad &\text{in } \Gamma(t).\\
# \end{align*}

# In[5]:


from dune.fem.space import lagrange
from dune.fem.scheme import galerkin

kspace = lagrange(igridView, dimRange=dim, order=1)
k  = TrialFunction(kspace)
kk = TestFunction(kspace)

curvature = kspace.interpolate([0]*dim, name="curvature")

ix = SpatialCoordinate(kspace)
C  = inner(k, kk) * dx
C -= inner(grad(ix), grad(kk)) * dx

kscheme = galerkin([C == 0])
res = kscheme.solve(curvature)

# ## Moving
#
# We will move the interface movement by evaluating the trace of the bulk velocity $u$.

# In[6]:


import numpy as np

x = SpatialCoordinate(pspace)
n = FacetNormal(pspace)
h = FacetArea(pspace)

def getShifts():
    mapper = hgrid.interfaceGrid.indexSet
    shifts = np.zeros((igridView.size(dim-1), dim))
    for e in igridView.elements:
        for v in e.subEntities(dim-1):
            x = e.geometry.toLocal(v.geometry.center)
            shifts[ mapper.index(v) ] = trace(uh)(e, x)
    return shifts

# ## Navier-Stokes equations
#
# We implement a splitting scheme similar to the one presented in [GBK20]_.
#
# .. [GBK20] J. Gerstenberger, S. Burbulla, D. Kröner. Discontinuous Galerkin method for incompressible two-phase flows. Finite Volumes for Complex Applications IX - Methods, Theoretical Aspects, Examples, pp. 675--683, 2020.

# In[7]:


from dune.mmesh import skeleton, interfaceIndicator
I = interfaceIndicator(igridView)

penu = Constant(1e6, name="penaltyu")
penp = Constant(1e6, name="penaltyp")

noslip = conditional(x[0] < 1e-6, 1, 0) + conditional(x[0] > 1-1e-6, 1, 0)

a1  = rho * inner(u - uh, v) / tau * dx
a1 += inner(grad(uh) * uh, v) * dx
a1 += inner(mu * grad(u), grad(v)) * dx

a1 += penu / h * inner(jump(u), jump(v)) * dS
a1 += dot(dot(avg(mu * grad(u)), n('+')), jump(v)) * dS
a1 += penu / h * inner(u - zero(dim), v) * noslip * ds
a1 += dot(dot(mu * grad(u), n), v) * noslip * ds

A1 = galerkin([a1 == 0], solver=("suitesparse", "umfpack"))


dirichlet = conditional(x[1] < 1e-6, 1, 0) + conditional(x[1] > 2-1e-6, 1, 0)
pD = conditional(x[1] < 1e-6, 1, 0)

a2  = inner(grad(p), grad(q)) * dx
a2 += penp / h * jump(p) * jump(q) * dS
a2 += dot(dot(avg(grad(p)), n('+')), jump(q)) * dS
a2 += penp / h * (p - pD) * q * dirichlet * ds
a2 += dot(dot(grad(p), n), q) * ds
a2 += inner(rho * div(uh1), q) / tau * dx

kappa = avg(skeleton(curvature))
a2 += penp / h * inner(sigma * kappa, n('+')) * jump(q) * I*dS

A2 = galerkin([a2 == 0], solver=("suitesparse", "umfpack"))


a3  = rho * inner(u - uh1, v) / tau * dx
a3 += inner(grad(ph), v) * dx

a3 += penu / h * inner(jump(u), jump(v)) * dS
a3 += penu / h * inner(u - zero(dim), v) * noslip * ds

A3 = galerkin([a3 == 0], solver=("suitesparse", "umfpack"))

# ## Timeloop
#
# Finally, we specify the time iteration where we adapt the mesh and solve the schemes.

# In[8]:


from dune.fem import parameter, adapt
parameter.append( { "fem.adaptation.method": "callback" } )
from dune.fem.plotting import plotPointData as plot
import matplotlib.pyplot as plt

N = 4
i = 0
fig, axs = plt.subplots(1, N, figsize=(16,6))

ph.interpolate(0)
uh.interpolate([0,0])
uh1.interpolate([0,0])

step = 0
t = 0
while t < T+dt:

    hgrid.moveInterface( dt*getShifts() )

    hgrid.markElements()
    adapt([ph, uh, uh1, dm])
    adapt([curvature])

    A1.solve(uh1)
    A2.solve(ph)
    A3.solve(uh)

    if int(N * t/T) > i:
        plot(ph, figure=(fig, axs[i]), gridLines='black', linewidth=0.02)
        plot(uh, figure=(fig, axs[i]), gridLines=None, vectors=[0,1])
        i += 1

    t += dt



from dune.grid import reader
from dune.mmesh import mmesh, monolithicSolve, skeleton, trace
from dune.mmesh.test.grids import line, plane
from dune.fem.space import lagrange
from dune.fem.scheme import galerkin
from dune.fem.function import integrate
from dune.ufl import DirichletBC, Constant
from ufl import *

# Let   Ω := (0, 1)^d,
#       γ := {x in Ω | x[d-1] = 0.5},
#     ∂ΩD := {x in ∂Ω | x[d-1] = 0 or x[d-1] = 1},
#     ∂ΩN := ∂Ω - ∂ΩD,
#     ∂γD := {x in ∂Ω | x[0] = 0 or x[0] = 1},
#     ∂γN := ∂γ - ∂γD,
#
# We use a finite element method to solve the following problem:
#          -Δu = 0  in Ω,
#    -Δλ + u*λ = g  in γ,
#            u = 0  on ∂ΩD,
#    grad(u)*n = 0  on ∂ΩN,
#            λ = 0  on ∂γD,
#    grad(λ)*n = 0  on ∂γN.

def coupledproblem(grid):
    dim = grid.dimension
    igrid = grid.hierarchicalGrid.interfaceGrid

    u_exact = lambda x : x[dim - 1]
    λ_exact = lambda x : sin(pi * x[0])
    ig = lambda x : (pi * pi + 0.5) * sin(pi * x[0])

    space = lagrange(grid, order=1)
    ispace = lagrange(igrid, order=1)

    x = SpatialCoordinate(space)
    u = TrialFunction(space)
    v = TestFunction(space)

    ix = SpatialCoordinate(ispace)
    λ = TrialFunction(ispace)
    μ = TestFunction(ispace)

    uh = space.interpolate(0., name="uh")
    λh = ispace.interpolate(0., name="lambdah")

    b = dot(grad(u), grad(v)) * dx
    dbc_top = DirichletBC(space, 0., x[dim - 1] <= 1e-8)
    dbc_btm = DirichletBC(space, 1., x[dim - 1] >= 1. - 1e-8)

    ib = dot(grad(λ), grad(μ)) * dx
    ib += trace(uh, igrid)('+') * λ * μ * dx
    il = ig(ix) * μ * dx
    idbc_lft = DirichletBC(ispace, 0., ix[0] <= 1e-8)
    idbc_rgt = DirichletBC(ispace, 0., ix[0] >= 1. - 1e-8)

    umfp = ("suitesparse", "umfpack")
    scheme = galerkin([b == 0., dbc_top, dbc_btm], solver=umfp)
    ischeme = galerkin([ib == il, idbc_lft, idbc_rgt], solver=umfp)
    monolithicSolve(schemes=(scheme, ischeme), targets=(uh, λh), verbose=1)

    grid.writeVTK(f"coupledsolve-{dim}d", pointdata={"uh": uh, "u_exact": u_exact(x)})
    igrid.writeVTK(f"coupledsolve-{dim}d-interface", pointdata={"λh": λh, "λ_exact": λ_exact(ix)})

    err = integrate(grid, abs(uh - u_exact(x)), order=5)
    ierr = integrate(igrid, abs(λh - λ_exact(ix)), order=5)

    assert(err < 1e-10)
    assert(ierr < 5e-3)

# 2D
grid = mmesh((reader.gmsh, line.filename), 2)
coupledproblem(grid)

# 3D
grid  = mmesh((reader.gmsh, plane.filename), 3)
coupledproblem(grid)



from dune.grid import reader
from dune.mmesh import mmesh
from dune.mmesh import distance
from ufl import grad
from dune.fem.function import integrate


# 2D
from dune.mmesh.test.grids import line
grid = mmesh((reader.gmsh, line.filename), 2)
igrid = grid.hierarchicalGrid.interfaceGrid

dist = distance(grid)

normal = grad(dist)
grid.writeVTK("distance-2d", pointdata={"d": dist, "n": [normal[0], normal[1], 0]}, nonconforming=True)
igrid.writeVTK("distance-2d-interface")

int = integrate(grid, dist, order=5)
nor = integrate(grid, normal, order=5)

assert(abs(int - 0.25) < 1e-8)
assert((nor * nor) < 1e-8)


# 3D
from dune.mmesh.test.grids import plane
grid  = mmesh((reader.gmsh, plane.filename), 3)
igrid = grid.hierarchicalGrid.interfaceGrid

dist = distance(grid)

normal = grad(dist)
grid.writeVTK("distance-3d", pointdata={"d": dist, "n": normal}, nonconforming=True)
igrid.writeVTK("distance-3d-interface")

from dune.fem.function import integrate
int = integrate(grid, dist, order=5)
nor = integrate(grid, normal, order=5)

assert(abs(int - 0.25) < 1e-3)
assert((nor * nor) < 1e-8)



#!/usr/bin/env python
# coding: utf-8

# # Manual adaptation
#
# This is an example of how to manually remove vertices and refine edges. With this, we can also implement anisotropic refinement.

# ## Grid creation
#
# We use the [horizontal](grids/horizontal.rst) grid file that contains an interface $\Gamma = [0.25, 0.75] \times {0.5}$ embedded in a domain $\Omega = [0,1]^2$.

# In[1]:


from dune.grid import reader
from dune.mmesh import mmesh

from dune.fem.view import adaptiveLeafGridView as adaptive
from dune.fem import parameter, adapt
parameter.append( { "fem.adaptation.method": "callback" } )

dim = 2
file = "grids/horizontal.msh"

gridView  = adaptive( mmesh((reader.gmsh, file), dim) )
hgrid = gridView.hierarchicalGrid

# ## Interpolating some high order function
#
# First, we interpolate some high order function.

# In[2]:


from ufl import *
from dune.fem.space import dglagrange as functionspace

space = functionspace(gridView, order=5)
x = SpatialCoordinate(space)

uh = space.interpolate( sin(100*x[0]), name="uh")
uh.plot()

# ## Removing and inserting vertices
#
# Then, we can insert and remove arbitrary vertices. We flag the vertices that should be removed and define points to be inserted within cells. Points inserted in cells will be connected to the all of the cell’s vertices. Afterwards, we simply call adapt.

# In[3]:


from dune.common import FieldVector

for v in gridView.vertices:
    if (v.geometry.center - FieldVector([0.3, 0.3])).two_norm < 0.1:
        hgrid.removeVertex(v)

for e in gridView.elements:
    if (e.geometry.center - FieldVector([0.7, 0.7])).two_norm < 0.2:
        hgrid.insertVertexInCell( e.geometry.center )

adapt([uh])
uh.plot()

# ## Refining edges
#
# Another possibility to change the grid is to bisect specific edges. Hereby, we have to define an edge for bisection by a corresponding intersection’s index.

# In[4]:


def myRefine():
    for e in gridView.elements:
        for i in gridView.intersections(e):
            if hgrid.isInterface(i):
                hgrid.refineEdge(e, i.indexInInside)

for i in range(5):
    myRefine()
    adapt([uh])

uh.plot()



name = "grid.msh"
h = 0.5
hf = 0.2

import gmsh
gmsh.initialize()
gmsh.option.setNumber("General.Verbosity", 0)
gmsh.option.setNumber("Mesh.MshFileVersion", 2.2)

gmsh.model.add(name)

p1 = gmsh.model.geo.addPoint(0, 0, 0, h, 1)
p2 = gmsh.model.geo.addPoint(3, 0, 0, h, 2)
p3 = gmsh.model.geo.addPoint(3, 3, 0, h, 3)
p4 = gmsh.model.geo.addPoint(0, 3, 0, h, 4)

l1 = gmsh.model.geo.addLine(p1, p2, 1)
l2 = gmsh.model.geo.addLine(p2, p3, 2)
l3 = gmsh.model.geo.addLine(p3, p4, 3)
l4 = gmsh.model.geo.addLine(p4, p1, 4)

p5 = gmsh.model.geo.addPoint(1, 1, 0, hf, 5)
p6 = gmsh.model.geo.addPoint(1.5, 1.5, 0, hf, 6)
p7 = gmsh.model.geo.addPoint(2, 2, 0, hf, 7)
p8 = gmsh.model.geo.addPoint(2, 1, 0, hf, 8)
lf1 = gmsh.model.geo.addLine(p5, p6, 10)
lf2 = gmsh.model.geo.addLine(p7, p6, 11)
lf3 = gmsh.model.geo.addLine(p8, p6, 12)

gmsh.model.geo.addCurveLoop([l1, l2, l3, l4], 1)
gmsh.model.geo.addPlaneSurface([1], 0)

gmsh.model.geo.synchronize()
gmsh.model.mesh.embed(1, [lf1, lf2, lf3], 2, 0)

gmsh.model.mesh.generate(dim=2)
gmsh.write(name)
gmsh.finalize()



import logging, traceback
logger = logging.getLogger(__name__)

from time import time
import numpy as np

# Return dof vector
def as_vector(df):
  try:
    return df.as_numpy
  except:
    try:
      return df.as_istl
    except:
      raise


def iterativeSolve(schemes, targets, callback=None, iter=100, tol=1e-8, f_tol=None, verbose=False, accelerate=False):
    """Helper function to solve bulk and interface scheme coupled iteratively.

    Args:
        schemes:  pair of schemes
        targets:  pair of discrete functions that should be solved for AND that are used in the coupling forms
        callback: update function that is called every time before solving a scheme
        iter:     maximum number of iterations
        tol:      objective tolerance between two iterates in two norm
        verbose:  print residuum for each iteration
        accelerate: use a vector formulation of Aitken's fix point acceleration proposed by Irons and Tuck.

    Returns:
        if converged and number of iterations

    Note:
        The targets also must be used in the coupling forms.
    """
    if f_tol is not None:
        print("f_tol is deprecated, use tol instead!")
        if tol != 1e-8:
            tol = f_tol

    assert len(schemes) == 2
    assert len(targets) == 2

    (A, B) = schemes
    (u, v) = targets

    a = u.copy()
    b = v.copy()

    if accelerate:
        Fa = a.copy()
        Fb = b.copy()

        FFa = a.copy()
        FFb = b.copy()

    def residuum(a, b):
        return np.sqrt(np.dot(a, a)+np.dot(b, b))

    if callback is not None:
        callback()

    converged = False
    for i in range(iter):
        a.assign(u)
        b.assign(v)

        # Evaluation: a, b -> Fa, Fb
        A.solve(u)
        B.solve(v)

        if accelerate:
            Fa.assign(u)
            Fb.assign(v)

            # Evaluation: Fa, Fb -> FFa, FFb
            if callback is not None:
                callback()
            A.solve(u)
            B.solve(v)
            FFa.assign(u)
            FFb.assign(v)

            # Irons-Tuck update
            DA  = as_vector(FFa) - as_vector(Fa)
            D2a = DA - as_vector(Fa) + as_vector(a)
            as_vector(u)[:] = as_vector(FFa)
            if np.dot(D2a, D2a) != 0:
                as_vector(u)[:] -= np.dot(DA, D2a) / np.dot(D2a, D2a) * DA

            DB  = as_vector(FFb) - as_vector(Fb)
            D2b  = DB - as_vector(Fb) + as_vector(b)
            as_vector(v)[:] = as_vector(FFb)
            if np.dot(D2b, D2b) != 0:
                as_vector(v)[:] -= np.dot(DB, D2b) / np.dot(D2b, D2b) * DB

        if callback is not None:
            callback()

        res = residuum(as_vector(u) - as_vector(a), as_vector(v) - as_vector(b))

        if verbose:
            print("{:3d}:".format(i), "[", "{:1.2e}".format(res), "]", flush=True)

        if res < tol:
            converged = True
            break

    return {'converged': converged, 'iterations': i}



def monolithicSolve(schemes, targets, callback=None, iter=30, tol=1e10, f_tol=1e-7, eps=1.49012e-8, verbose=0, iterative=False):
    """Helper function to solve bulk and interface scheme coupled monolithically.
       A newton method assembling the underlying jacobian matrix.
       The coupling jacobian blocks are evaluated by finite differences.
       We provide an implementation with a fast C++ backend.

    Args:
        schemes:  pair of schemes
        targets:  pair of discrete functions that should be solved for AND that are used in the coupling forms
        callback: update function that is called every time before solving a scheme
        iter:     maximum number of iterations
        tol:      objective residual of iteration step in two norm
        f_tol:    objective residual of function value in two norm
        eps:      step size for finite difference
        verbose:  1: print residuum for each newton iteration, 2: print details
        iterative: Use the experimental iterative solver backend instead of UMFPack. Remark that the solver arguments of the bulk scheme are taken and preconditioning is not supported yet.

    Returns:
        if converged
    """
    assert len(schemes) == 2
    assert len(targets) == 2

    (scheme, ischeme) = schemes
    (uh, th) = targets

    n = len(as_vector(uh))
    m = len(as_vector(th))

    def call():
        if callback is not None:
            callback()

    from numpy.linalg import norm

    # Evaluate
    f = uh.copy()
    g = th.copy()

    call()
    scheme(uh, f)
    ischeme(th, g)

    # Load C++ jacobian implementation
    import hashlib
    from dune.generator import Constructor
    from dune.generator.generator import SimpleGenerator
    from mpi4py import MPI
    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()

    if iterative:
      header = "jacobian_iterative.hh"
    else:
      header = "jacobian.hh"
    typeName = "Dune::Python::MMesh::Jacobian< " + scheme.cppTypeName + ", " \
        + ischeme.cppTypeName + ", " + uh.cppTypeName + ", " + th.cppTypeName + " >"
    includes = scheme.cppIncludes + ischeme.cppIncludes + ["dune/python/mmesh/"+header]
    moduleName = "jacobian_" + hashlib.md5(typeName.encode('utf8')).hexdigest()
    constructor = Constructor(['const '+scheme.cppTypeName+'& scheme','const '+ischeme.cppTypeName+' &ischeme', 'const '+uh.cppTypeName+' &uh', 'const '+th.cppTypeName+' &th', 'const double eps', 'const std::function<void()> &callback'],
                              ['return new ' + typeName + '( scheme, ischeme, uh, th, eps, callback );'],
                              ['pybind11::keep_alive< 1, 2 >()', 'pybind11::keep_alive< 1, 3 >()', 'pybind11::keep_alive< 1, 4 >()', 'pybind11::keep_alive< 1, 6 >()'])

    generator = SimpleGenerator("Jacobian", "Dune::Python::MMesh")
    module = generator.load(includes, typeName, moduleName, constructor)
    jacobian = module.Jacobian(scheme, ischeme, uh, th, eps, call)
    jacobian.init()

    ux = uh.copy()
    tx = th.copy()

    def norm(u, t):
      return np.sqrt(u.scalarProductDofs(u) + t.scalarProductDofs(t))

    for i in range(1, iter+1):

        jacobian.update(uh, th)


        solveTime = -time()
        jacobian.solve(f, g, ux, tx)
        solveTime += time()

        uh -= ux
        th -= tx
        xres = norm(ux, tx)

        call()
        scheme(uh, f)
        ischeme(th, g)

        fres = norm(f, g)

        if verbose > 0 and rank == 0:
            print(" i:", i, " |Δx| =", "{:1.8e}".format(xres), "",  "|f| =", "{:1.8e}".format(fres))
            if verbose > 1:
                print(f"Solve took {solveTime:.6f} seconds.\n")
                file = open('runtime.txt', 'w')
                file.write(str(solveTime))
                file.close()

        if xres < tol and fres < f_tol:
            return True

    return False



filename = "plane.msh"
h = 0.2
hf = 0.05

import gmsh
gmsh.initialize()
gmsh.option.setNumber("General.Verbosity", 0)
gmsh.option.setNumber("Mesh.MshFileVersion", 2.2)

gmsh.model.add(filename)

occ = gmsh.model.occ
box = occ.addBox(0, 0, 0, 1, 1, 1)
boxPoints = gmsh.model.occ.getEntities(0)

plane = occ.addRectangle(0, 0, 0.5, 1, 1)
planePoints = gmsh.model.occ.getEntities(0)

occ.mesh.setSize(planePoints, hf)
occ.mesh.setSize(boxPoints, h)

occ.fragment([(3, box)], [(2, plane)])
occ.synchronize()

gmsh.model.mesh.generate(dim=3)
gmsh.write(filename)
gmsh.finalize()



"""
.. module:: _grids
"""

from __future__ import absolute_import, division, print_function, unicode_literals

import sys
import logging
logger = logging.getLogger(__name__)

def mmesh(constructor, dimgrid=None, **parameters):
    """Create an MMesh grid.

    Args:
        constructor: Grid constructor, e.g. (dune.grid.reader.gmsh, 'grid.msh') or dune.grid.cartesianDomain([0,0],[1,1],[10,10]).
        dimgrid (int, optional): The world dimension (must be 2 or 3). Might be guessed by constructor.
        **parameters: Additional parameters.

    Returns:
        An MMesh grid instance.

    Examples:

        Creating simple structured grid::

            from dune.grid import cartesianDomain
            from dune.mmesh import mmesh
            grid = mmesh(cartesianDomain([0,0], [1,1], [10,10]))


        Using a .msh file::

            from dune.grid import reader
            from dune.mmesh import mmesh
            grid = mmesh((reader.gmsh, "grid.msh"), 2)
            igrid = grid.hierarchicalGrid.interfaceGrid


        where grid.msh can be generated by `gmsh <https://pypi.org/project/gmsh/>`_ [GR09]_. For instance::

            name = "grid.msh"
            h = 0.1
            hf = 0.01

            import gmsh
            gmsh.initialize()
            gmsh.option.setNumber("Mesh.MshFileVersion", 2.2)

            gmsh.model.add("name")

            geo = gmsh.model.geo

            p1 = geo.addPoint(0, 0, 0, h, 1)
            p2 = geo.addPoint(1, 0, 0, h, 2)
            p3 = geo.addPoint(1, 1, 0, h, 3)
            p4 = geo.addPoint(0, 1, 0, h, 4)

            l1 = geo.addLine(p1, p2, 1)
            l2 = geo.addLine(p2, p3, 2)
            l3 = geo.addLine(p3, p4, 3)
            l4 = geo.addLine(p4, p1, 4)

            p5 = geo.addPoint(0.25, 0.5, 0, hf, 5)
            p6 = geo.addPoint(0.75, 0.5, 0, hf, 6)
            lf = geo.addLine(p5, p6, 10)

            geo.addCurveLoop([l1, l2, l3, l4], 1)
            geo.addPlaneSurface([1], 0)

            geo.synchronize()

            gmsh.model.mesh.embed(1, [lf], 2, 0)
            gmsh.model.mesh.generate(dim=2)
            gmsh.write(name)
            gmsh.finalize()

        .. [GR09] C. Geuzaine and J.-F. Remacle. Gmsh: a three-dimensional finite element mesh generator with built-in pre- and post-processing facilities. International Journal for Numerical Methods in Engineering 79(11), pp. 1309-1331, 2009.

    """

    if not dimgrid:
        from dune.grid.grid_generator import getDimgrid
        dimgrid = getDimgrid(constructor)

    if not (2 <= dimgrid and dimgrid <= 3):
        raise KeyError("Parameter error in MMesh with dimgrid=" + str(dimgrid) + ": dimgrid has to be either 2 or 3")

    if dimgrid == 2:
        import dune.mmesh._grid2d as gridModule
        import dune.mmesh._interfacegrid2d as igridModule
    elif dimgrid == 3:
        import dune.mmesh._grid3d as gridModule
        import dune.mmesh._interfacegrid3d as igridModule

    gridView = gridModule.LeafGrid(gridModule.reader(constructor))


    # in case of a cartesian domain store if old or new boundary ids was used
    # this can be removed in later version - it is only used in dune-fem
    # to give a warning that the boundary ids for the cartesian domains have changed
    try:
        gridView.hierarchicalGrid._cartesianConstructionWithIds = constructor.boundaryWasSet
    except AttributeError:
        pass

    interfaceGrid = lambda self: self.interfaceHierarchicalGrid.leafView
    gridModule.HierarchicalGrid.interfaceGrid = property(interfaceGrid)

    return gridView

grid_registry = {
    "MMesh" : mmesh
}

if __name__ == "__main__":
    import doctest
    doctest.testmod(optionflags=doctest.ELLIPSIS)



from dune.grid import yaspGrid, reader, cartesianDomain
from dune.alugrid import aluConformGrid as alu
from dune.mmesh import mmesh
from dune.fem.space import lagrange
from dune.fem.scheme import galerkin as scheme
from dune.fem.operator import galerkin as op
from dune.ufl import DirichletBC
from ufl import *

def warn(grid):
    try:
        return grid.hierarchicalGrid._cartesianConstructionWithIds
    except AttributeError:
        return False

def test(lgv, domain):
    grid = lgv(domain, dimgrid=2)
    print("if only id is '1' need warning?\t\t", warn(grid))
    spc = lagrange(grid)
    u = TrialFunction(spc)
    v = TestFunction(spc)
    eqn = u * v * dx == 0
    dbc = DirichletBC(spc, 0)
    dbc1 = DirichletBC(spc, 0, 1)
    dbc2 = DirichletBC(spc, 0, 2)
    print("only id=1:")
    scheme([eqn, dbc1])
    print("no id:")
    scheme([eqn, dbc])
    print("id=1 and 2:")
    scheme([eqn, dbc1, dbc2])

print("alu, old boundary treatment", end=" -> ")
test( alu, cartesianDomain([0,0], [1,1], [10,10], boundary=False) )
print("alu, new boundary treatment", end=" -> ")
test( alu, cartesianDomain([0,0], [1,1], [10,10], boundary=True) )
print("alu, default treatment", end=" -> ")
test( alu, cartesianDomain([0,0], [1,1], [10,10]) )

print("yasp, old boundary treatment", end=" -> ")
test( yaspGrid, cartesianDomain([0,0], [1,1], [10,10], boundary=False) )
print("yasp, new boundary treatment", end=" -> ")
test( yaspGrid, cartesianDomain([0,0], [1,1], [10,10], boundary=True) )
print("yasp, default treatment", end=" -> ")
test( yaspGrid, cartesianDomain([0,0], [1,1], [10,10]) )

print("mmesh, old boundary treatment", end=" -> ")
test( mmesh, cartesianDomain([0,0], [1,1], [10,10], boundary=False) )
print("mmesh, new boundary treatment", end=" -> ")
test( mmesh, cartesianDomain([0,0], [1,1], [10,10], boundary=True) )
print("mmesh, default treatment", end=" -> ")
test( mmesh, cartesianDomain([0,0], [1,1], [10,10]) )



from dune.grid import reader
from dune.mmesh import mmesh
from dune.fem.space import dglagrange
from dune.fem.function import integrate
from ufl import *
from dune.ufl import Constant
from dune.fem.scheme import galerkin
from time import time
from mpi4py import MPI
rank = MPI.COMM_WORLD.Get_rank()

def algorithm(grid, name):
  print(grid.size(0), "cells,", grid.ghostSize(0), "ghost")
  space = dglagrange(grid, order=1, storage="istl")
  x = SpatialCoordinate(space)
  uh = space.interpolate(0, name="uh")

  u = TrialFunction(space)
  v = TestFunction(space)

  exact = sin(pi*x[0]) * sin(pi*x[1])
  beta = Constant(3.01, name="beta")
  n = FacetNormal(space)
  h = avg(FacetArea(space))
  hBnd = FacetArea(space)

  A = inner(grad(u), grad(v)) * dx
  A += beta / h * inner(jump(u), jump(v)) * dS
  A -= dot(dot(avg(grad(u)), n('+')), jump(v)) * dS
  A -= dot(dot(avg(grad(v)), n('+')), jump(u)) * dS
  A += beta / hBnd * inner(u - exact, v) * ds
  A -= dot(dot(grad(u), n), v) * ds
  A -= dot(dot(grad(v), n), u - exact) * ds

  if name == "interface":
    b = -exact.dx(0).dx(0) * v * dx
  else:
    b = -div( grad(exact) ) * v * dx

  scheme = galerkin([A == b], solver='cg', parameters={"newton.linear.verbose": "false"})

  took = -time()
  scheme.solve(uh)
  took += time()

  if rank == 0:
    print(f"Took {took:.6f}")

  grid.writeVTK("laplace-"+name, pointdata={"uh": uh, "exact": exact}, nonconforming=True)

  l2 = integrate(grid, sqrt(dot(uh-exact, uh-exact)), order=5)
  if(l2 > 1e-2):
    print(l2)
    raise


########
# MAIN #
########


from dune.mmesh.test.grids import line, vertical

for file in [line.filename, vertical.filename]:
  grid = mmesh((reader.gmsh, file), 2)
  hgrid = grid.hierarchicalGrid
  igrid = hgrid.interfaceGrid

  # Run bulk
  algorithm(grid, "bulk")

  # Run interface
  algorithm(igrid, "interface")



#!/usr/bin/env python
# coding: utf-8

# # Coupling to the interface
#
# This is an example of how to use Dune-MMesh and solve coupled problems on the bulk and interface grid.

# ## Grid creation
#
# We use the [horizontal](grids/horizontal.rst) grid file that contains an interface $\Gamma = [0.25, 0.75] \times {0.5}$ embedded in a domain $\Omega = [0,1]^2$. Grid creation from a mesh file works as follow.

# In[1]:


from dune.grid import reader
from dune.mmesh import mmesh
dim = 2
file = "grids/horizontal.msh"
gridView  = mmesh((reader.gmsh, file), dim)
igridView = gridView.hierarchicalGrid.interfaceGrid

# ## Solve a problem on the bulk grid

# Let us solve the Poisson equation
#
# \begin{align}
# -\Delta u = f & \qquad \text{in}\ \Omega
# \end{align}
#
# on the bulk grid. We use the manufactured solution $\hat u = \sin(4 \pi x y)$ and therefore apply the source term $f = -\operatorname{div}( \nabla \hat u )$. The weak form of the problem above reads
# \begin{align}
# \int_\Omega \nabla u \cdot \nabla v~dx &= \int_\Omega f v~dx
# \end{align}
# for all corresponding test functions $v$. This can be implemented as follows.

# In[2]:


from ufl import *
from dune.ufl import DirichletBC
from dune.fem.space import lagrange
from dune.fem.scheme import galerkin
from dune.fem.function import integrate

space = lagrange(gridView, order=3)
u = TrialFunction(space)
v = TestFunction(space)

x = SpatialCoordinate(space)
exact = sin(x[0]*x[1]*4*pi)
f = -div(grad(exact))

a = inner(grad(u), grad(v)) * dx
b = f * v * dx

scheme = galerkin([a == b, DirichletBC(space, exact)], solver=("suitesparse", "umfpack"))
uh = space.interpolate(0, name="solution")
scheme.solve(target=uh)

def L2(u1, u2):
    return sqrt(integrate(u1.grid, (u1-u2)**2, order=5))

L2(uh, exact)

# ## Solve a problem on the interface
#
# We can solve similar problem on the interface $\Gamma$ like
# \begin{align}
# -\Delta u_\Gamma = f & \qquad \text{in}\ \Gamma
# \end{align}
# with the weak form
# \begin{align}
# \int_\Gamma \nabla u_\Gamma \cdot \nabla v_\Gamma~dx &= \int_\Gamma f v_\Gamma~dx
# \end{align}
# for all corresponding test functions $v_\Gamma$.

# In[3]:


ispace = lagrange(igridView, order=3)
iuh = ispace.interpolate(0, name="isolution")

iu = TrialFunction(ispace)
iv = TestFunction(ispace)

ix = SpatialCoordinate(ispace)
iexact = sin(0.5*ix[dim-2]*4*pi)
iF = -div(grad(iexact))

ia = inner(grad(iu), grad(iv)) * dx
ib = iF * iv * dx

ischeme = galerkin([ia == ib, DirichletBC(ispace, iexact)])
ischeme.solve(target=iuh)
L2(iuh, iexact)

# We can use the `plotPointData` function to visualize the solution of both grids.

# In[4]:


import matplotlib.pyplot as plt
from dune.fem.plotting import plotPointData as plot

figure = plt.figure(figsize=(3,3))
plot(uh, figure=figure, gridLines=None)
plot(iuh, figure=figure, linewidth=0.04, colorbar=None)

# ## Couple bulk to surface
#
# Dune-MMesh makes it possible to compute traces of discrete functions on $\Omega$ along $\Gamma$.

# In[5]:


from dune.mmesh import trace
tr = avg(trace(uh))
ib = inner(grad(tr), grad(iv)) * dx

iuh.interpolate(0)
ischeme = galerkin([ia == ib, DirichletBC(ispace, avg(trace(uh)))])
ischeme.solve(target=iuh)
L2(iuh, iexact)

# ## Couple surface to bulk
#
# Similarly, we can evaluate a discrete function on $\Gamma$ at the skeleton of the triangulation of $\Omega$.

# In[6]:


from dune.mmesh import skeleton

sk = skeleton(iuh)
b = avg(sk) * avg(v) * dS

uh.interpolate(0)
scheme = galerkin([a == b, DirichletBC(space, 0)])
scheme.solve(target=uh)

figure = plt.figure(figsize=(3,3))
plot(uh, figure=figure, gridLines=None)

# ## Compute jump of gradient of traces
#
# We can use trace and skeleton within UFL expressions. This is useful when using jump and average operators, but also to compute gradients and more. Together with the utility function `normals`, which returns the positive normal of the underlying bulk facet, we can determine the orientation of the restricted values.

# In[7]:


from dune.mmesh import normals
inormal = normals(igridView)
jmp = jump(grad(trace(uh)), inormal)



from time import time

from dune.mmesh.test.grids import tjunction

from dune.grid import reader
from dune.mmesh import mmesh
gridView  = mmesh((reader.gmsh, tjunction.name), 2)
igridView = gridView.hierarchicalGrid.interfaceGrid

from ufl import *
from dune.fem.space import lagrange, dglagrange
space = dglagrange(gridView, order=1)
ispace = dglagrange(igridView, order=1)
u = TrialFunction(space)
v = TestFunction(space)
iu = TrialFunction(ispace)
iv = TestFunction(ispace)
uh = space.interpolate(0, name="uh")
iuh = ispace.interpolate(0, name="iuh")

from dune.ufl import Constant
from dune.mmesh import interfaceIndicator
q     = Constant(1, name="q")
omega = Constant(1e-6, name="omega")
beta  = Constant(1, name="beta")
n = FacetNormal(space)
ni = FacetNormal(ispace)
I = interfaceIndicator(igridView)

a  = inner(grad(u), grad(v)) * dx
a += beta * inner(jump(u), jump(v)) * (1-I)*dS
a -= dot(dot(avg(grad(u)), n('+')), jump(v)) * (1-I)*dS
a += beta * inner(u - 0, v) * ds
a -= dot(dot(grad(u), n), v) * ds

ia  = inner(grad(iu), grad(iv)) * dx
ia += beta * inner(jump(iu), jump(iv)) * dS
ia -= dot(dot(avg(grad(iu)), ni('+')), jump(iv)) * dS
ib  = q * iv * dx

from dune.mmesh import skeleton, trace
omega = Constant(1e-6, name="omega")
a -= (skeleton(iuh)('+') - u('+')) / omega * v('+') * I*dS
a -= (skeleton(iuh)('-') - u('-')) / omega * v('-') * I*dS
ia += (iu - trace(uh)('+')) / omega * iv * dx
ia += (iu - trace(uh)('-')) / omega * iv * dx

from dune.fem.scheme import galerkin
scheme  = galerkin([a == 0])
ischeme = galerkin([ia == ib])
from dune.mmesh import monolithicSolve
dt = -time()
monolithicSolve(schemes=(scheme, ischeme), targets=(uh, iuh), verbose=True)
dt += time()
print(f"Took {dt:.6f}")

from dune.fem.function import integrate
intBulk = integrate(gridView, uh, order=1)
intInterface = integrate(igridView, iuh, order=1)

print(intBulk, intInterface)
assert(abs(intBulk - 0.066) < 1e-3)
assert(abs(intInterface - 0.192) < 1e-3)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import sys
import subprocess
sys.path.insert(0, os.path.abspath('../python/'))

# -- Project information -----------------------------------------------------

project = 'dune-mmesh'
copyright = '2021, Samuel Burbulla'
author = 'Samuel Burbulla, Andreas Dedner, Maximilian Hörl, Christian Rohde'

# The full version, including alpha/beta/rc tags
release = 'master'


# -- General configuration ---------------------------------------------------

autodoc_mock_imports = ["dune.generator", "dune.fem"]

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
  "sphinx.ext.intersphinx",
  "sphinx.ext.napoleon",
  "sphinx.ext.autodoc",
  "sphinx.ext.mathjax",
  "sphinx.ext.viewcode",
  "sphinx_rtd_theme",
  "breathe",
  "nbsphinx",
  "sphinxcontrib.tikz",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

breathe_projects = { "dune-mmesh": "xml" }
breathe_default_project = 'dune-mmesh'

tikz_proc_suite = 'GhostScript'
tikz_tikzlibraries = 'calc'
tikz_resolution = 1000

nbsphinx_execute = 'never'

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.

html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

# -- Options for Latex output -------------------------------------------------
latex_documents = [
    ('index', 'dune-mmesh.tex', 'Dune-MMesh: The DUNE Grid Module for Moving Interfaces', author, 'article')
]

latex_toplevel_sectioning = 'section'
latex_elements = {
    'tableofcontents': '',
    'preamble': r'''
\newcommand{\jump}[1]{[\mskip-5mu[ #1 ]\mskip-5mu]}
\newcommand{\avg}[1]{\{\mskip-5mu\{ #1 \}\mskip-5mu\}}
''',
    'printindex': '',
}

def generate_doxygen_xml(app):
    subprocess.call(["doxygen", "doxygen/Doxyfile"], cwd=app.confdir)

def setup(app):
    app.connect("builder-inited", generate_doxygen_xml)



import logging, traceback
logger = logging.getLogger(__name__)

import hashlib

from dune.generator.generator import SimpleGenerator

def skeleton(interfaceFunction, grid=None):
    """Return the skeleton representation of a discrete function on the interface grid.

    Args:
        interfaceFunction: The discrete function on the interface grid.
        grid (Grid, optional): The bulk grid. Necessary, if wrapped.

    Returns:
        Skeleton representation of given interface function.

    Note:
        This function has to be restricted when evaluated on facets, e.g. using avg(skeleton).
    """
    if grid == None:
        grid = interfaceFunction.space.gridView.hierarchicalGrid.bulkGrid

    includes = ["dune/python/mmesh/pyskeletontrace.hh"]
    includes += interfaceFunction.cppIncludes + grid.cppIncludes
    generator = SimpleGenerator("SkeletonGF", "Dune::Fem")

    typeName = "Dune::Fem::SkeletonGF< " + grid.cppTypeName + ", " + interfaceFunction.cppTypeName + " >"
    moduleName = "skeleton_" + hashlib.md5(typeName.encode('utf8')).hexdigest()
    cls = generator.load(includes, typeName, moduleName)
    skeleton = cls.SkeletonGF(grid, interfaceFunction)

    import dune.ufl
    skeleton = dune.ufl.GridFunction(skeleton)

    interfaceFunction.skeleton = skeleton
    return skeleton


def trace(bulkFunction, igrid=None, restrictTo=None):
    """Return the trace representation of a discrete function on the bulk grid.

    Args:
        bulkFunction: The discrete function on the bulk grid.
        igrid (InterfaceGrid, optional): The interface grid.
        restrictTo: already restrict the trace to '+' or '-' side

    Returns:
        Trace representation of a given interface function.
        This function has to be restricted to positive ('+') or negative side ('-').
    """
    if igrid == None:
      igrid = bulkFunction.space.gridView.hierarchicalGrid.interfaceGrid

    traces = {}
    includes = ["dune/python/mmesh/pyskeletontrace.hh"]
    includes += bulkFunction.cppIncludes + igrid.cppIncludes
    generator = SimpleGenerator(["TraceGF","TraceGF"], "Dune::Fem", pythonname=["TraceGFP","TraceGFM"])
    typeName = []
    for side in [ "in", "out" ]:
        sideStr = "Dune::Fem::IntersectionSide::" + side
        typeName += ["Dune::Fem::TraceGF< " + igrid.cppTypeName + ", " + bulkFunction.cppTypeName + ", " + sideStr + " >"]
    moduleName = "skeleton_"+hashlib.md5(''.join(typeName).encode('utf8')).hexdigest()
    module = generator.load(includes, typeName, moduleName)
    traces["in"]  = module.TraceGFP(igrid, bulkFunction)
    traces["out"] = module.TraceGFM(igrid, bulkFunction)

    import ufl
    import dune.ufl
    trace_p = dune.ufl.GridFunction(traces["in"])
    trace_m = dune.ufl.GridFunction(traces["out"])

    if restrictTo is not None:
        if not restrictTo in ['+', '-']:
            raise Exception("restrictTo must be either '+' or '-'")
        return trace_p if restrictTo == '+' else trace_m

    if bulkFunction.scalar:
        trace_p = trace_p.toVectorCoefficient()
        trace_m = trace_m.toVectorCoefficient()

    trace = trace_p
    predefined = {}
    predefined[trace('+')]                     = trace_p
    predefined[trace('-')]                     = trace_m
    predefined[ufl.grad(trace)('+')]           = ufl.grad(trace_p)
    predefined[ufl.grad(trace)('-')]           = ufl.grad(trace_m)
    predefined[ufl.grad(ufl.grad(trace))('+')] = ufl.grad(ufl.grad(trace_p))
    predefined[ufl.grad(ufl.grad(trace))('-')] = ufl.grad(ufl.grad(trace_m))
    trace.predefined = predefined
    if bulkFunction.scalar:
        trace = trace[0]

    bulkFunction.trace = trace
    return trace
################################################################################



from dune.grid import reader
from dune.mmesh import mmesh
from dune.mmesh import normals
from dune.fem.function import integrate


# 2D
from dune.mmesh.test.grids import line
grid = mmesh((reader.gmsh, line.filename), 2)
igrid = grid.hierarchicalGrid.interfaceGrid

normal = normals(igrid)

igrid.writeVTK("normals-2d", pointdata={"n": normal})

int = integrate(igrid, [abs(normal[0]), abs(normal[1])], order=0)
assert(abs(int[0]) < 1e-8)
assert(abs(int[1] - 1.) < 1e-8)


# 3D
from dune.mmesh.test.grids import plane
grid  = mmesh((reader.gmsh, plane.filename), 3)
igrid = grid.hierarchicalGrid.interfaceGrid

normal = normals(igrid)

igrid.writeVTK("normals-3d", pointdata={"n": normal})

int = integrate(igrid, [abs(normal[0]), abs(normal[1]), abs(normal[2])], order=0)
assert(abs(int[0]) < 1e-8)
assert(abs(int[1]) < 1e-8)
assert(abs(int[2] - 1.) < 1e-8)



import logging, traceback
logger = logging.getLogger(__name__)

import io
from dune.generator import algorithm, builder
from dune.common.hashit import hashIt
from dune.generator.exceptions import CompileError

def interfaceIndicator(igrid, grid=None, restrict=True):
    """Return indicator of interface edges.

    Args:
        igrid: The interfacegrid.
        grid: The bulk grid. Pass this parameter if the bulk grid is wrapped, e.g., by a geometryGridView.
        restrict (bool, optional): If True, the returned UFL expression is restricted.

    Returns:
        Skeleton function that is one at interface edges.
    """

    # Pre-compiled version
    if grid is None:
        grid = igrid.hierarchicalGrid.bulkGrid
        if igrid.dimension == 1:
            import dune.mmesh._utility2d as module
        else:
            import dune.mmesh._utility3d as module

    # JIT version for wrapped grid
    else:
        moduleName = "interfaceindicator_" + hashIt(grid.cppTypeName)
        signature = "Compiling InterfaceIndicator"
        source = """
#include <config.h>
#include <dune/mmesh/mmesh.hh>
#include <dune/python/mmesh/utility.hh>
#include <dune/python/grid/hierarchical.hh>
#include <dune/python/common/typeregistry.hh>
#include <dune/python/pybind11/pybind11.h>
#include <dune/python/pybind11/stl.h>
"""
        includes = sorted(set(grid.cppIncludes))
        source += "".join(["#include <" + i + ">\n" for i in includes])
        source += """
PYBIND11_MODULE( """ + moduleName + """, module )
{
  // InterfaceIndicator
  auto clsIndicator = Dune::Python::insertClass< Dune::Fem::InterfaceIndicator< typename """ + grid.cppTypeName + """ > >(
    module,
    "InterfaceIndicator",
    Dune::Python::GenerateTypeName("Dune::Fem::InterfaceIndicator< typename """ + grid.cppTypeName + """ >"),
    Dune::Python::IncludeFiles{"dune/mmesh/mmesh.hh", "dune/python/grid/hierarchical.hh", "dune/python/mmesh/utility.hh"}
  ).first;
  Dune::Fem::registerInterfaceIndicator( module, clsIndicator );
}
"""
        module = builder.load(moduleName, source, signature)

    indicator = module.InterfaceIndicator(grid)
    from dune.ufl import GridFunction
    indicator = GridFunction(indicator)

    if restrict:
        from ufl import avg
        return avg(indicator)
    else:
        return indicator


def normals(igrid):
    """Return normal vectors to the interface grid elements.

    Args:
        igrid: The interface grid.

    Returns:
        Grid function on the interface grid. Coincides with n('+') of the bulk facet normal.
    """
    igrid = igrid.hierarchicalGrid.leafView

    if igrid.dimension == 1:
        import dune.mmesh._utility2d as module
    else:
        import dune.mmesh._utility3d as module

    normals = module.Normals(igrid)
    from dune.ufl import GridFunction
    normals = GridFunction(normals)
    return normals


def distance(grid):
    """Return function representing the distance to the interface

    Args:
        grid: The grid.

    Returns:
        Piecewise linear grid function.
    """

    if grid.dimension == 2:
        import dune.mmesh._utility2d as module
    else:
        import dune.mmesh._utility3d as module

    distance = module.Distance(grid.hierarchicalGrid.leafView)
    from dune.ufl import GridFunction
    distance = GridFunction(distance)
    return distance


def domainMarker(grid, wrapped=False):
    """Return domain markers passed by .msh grid file.

    Args:
       grid: The grid.
       wrapped: Set this to true if grid is wrapped.

    Returns:
       Grid function with cell-wise markers from .msh file.
    """
    code = """
    #include <functional>
    template <class GV>
    auto domainMarker(const GV &gv) {
      auto ret = [] (const auto& entity, const auto& xLocal) mutable -> auto {
        return entity.impl().domainMarker();
      };
      return ret;
    }
    """
    from dune.fem.space import finiteVolume
    from dune.fem.function import cppFunction
    if not wrapped:
        cppfunc = cppFunction(grid, name="domainMarker", order=0, fctName="domainMarker", includes=io.StringIO(code), args=[grid])
    else:
        code = code.replace("impl()", "impl().hostEntity().impl()")
        cppfunc = cppFunction(grid, name="domainMarkerImpl", order=0, fctName="domainMarker", includes=io.StringIO(code), args=[grid])
    fvspace = finiteVolume(grid)
    return fvspace.interpolate(cppfunc, name="domainMarker")


def interfaceDomainMarker(igrid):
    """Return interface domain markers passed by .msh grid file.

    Args:
       igrid: The interface grid.

    Returns:
       Interface grid function with cell-wise markers from .msh file.
    """
    code = """
    #include <iostream>
    #include <functional>
    template <class GV>
    auto interfaceDomainMarker(const GV &igv) {
      auto ret = [&igv] (const auto& entity, const auto& xLocal) mutable -> auto {
        return igv.grid().domainMarker(entity);
      };
      return ret;
    }
    """
    from dune.fem.function import cppFunction
    return cppFunction(igrid, name="interfaceDomainMarker", order=0, fctName="interfaceDomainMarker", includes=io.StringIO(code), args=[igrid])


def edgeMovement(grid, shifts):
    """Return linear interpolation of vertex shifts.

    Args:
       grid: The grid.
       shifts:   List of shifts.

    Returns:
       Vector-valued grid function.
    """
    code="""
    #include <functional>
    #include <dune/geometry/multilineargeometry.hh>
    #include <dune/python/pybind11/pybind11.h>
    #include <dune/python/pybind11/numpy.h>

    template <class GV>
    auto edgeMovement(const GV &gv, const pybind11::array_t<double>& input) {
        static constexpr int dim = GV::dimension;
        using GlobalCoordinate = Dune::FieldVector<double, dim>;

        // obtain shifts from buffer protocol
        pybind11::buffer_info buffer = input.request();
        double *ptr = (double *) buffer.ptr;
        int n = buffer.shape[0];

        std::vector<GlobalCoordinate> shifts( n );
        for ( std::size_t i = 0; i < n; ++i )
          for ( int d = 0; d < dim; ++d )
            shifts[i][d] = ptr[dim*i+d];

        const auto& igrid = gv.grid().interfaceGrid();
        const auto& iindexSet = igrid.leafIndexSet();

        auto ret = [shifts, &igrid, &iindexSet] (const auto& entity, const auto& xLocal) mutable -> auto {
            // return a linear interpolation of the vertex shift values
            std::vector<GlobalCoordinate> shift( dim+1 );
            for ( std::size_t i = 0; i < entity.subEntities( dim ); ++i )
            {
                const auto& vertex = entity.template subEntity<dim>( i );
                if ( vertex.impl().isInterface() )
                {
                    // cast to interface grid vertex to obtain index
                    const auto ivertex = igrid.entity( vertex.impl().hostEntity() );
                    shift[i] = shifts[ iindexSet.index( ivertex ) ];
                }
            }
            const Dune::MultiLinearGeometry<double, dim, dim> interpolation(entity.type(), shift);
            return interpolation.global(xLocal);
        };
        return ret;
    }
    """
    from dune.fem.function import cppFunction
    cppFunc = cppFunction(grid, name="edgeMovement", order=1, fctName="edgeMovement", includes=io.StringIO(code), args=[grid, shifts])
    from dune.fem.space import lagrange
    space = lagrange(grid, dimRange=grid.dimWorld, order=1)
    return space.interpolate(cppFunc, name="edgeMovement")


def interfaceEdgeMovement(igrid, shifts):
    """Return linear interpolation of vertex shifts on the interface.

    Args:
       igrid: The interface grid.
       shifts:    List of shifts.

    Returns:
       Vector-valued interface grid function.
    """
    code="""
    #include <functional>
    #include <dune/geometry/multilineargeometry.hh>
    #include <dune/python/pybind11/pybind11.h>
    #include <dune/python/pybind11/numpy.h>

    template <class GV>
    auto interfaceEdgeMovement(const GV &gv, const pybind11::array_t<double>& input) {
        static constexpr int dim = GV::dimension;
        static constexpr int dimw = GV::dimensionworld;
        using GlobalCoordinate = Dune::FieldVector<double, dimw>;

        // obtain shifts from buffer protocol
        pybind11::buffer_info buffer = input.request();
        double *ptr = (double *) buffer.ptr;
        int n = buffer.shape[0];

        std::vector<GlobalCoordinate> shifts( n );
        for ( std::size_t i = 0; i < n; ++i )
          for ( int d = 0; d < dimw; ++d )
            shifts[i][d] = ptr[dimw*i+d];

        const auto& igrid = gv.grid();
        const auto& iindexSet = igrid.leafIndexSet();

        auto ret = [shifts, &igrid, &iindexSet] (const auto& entity, const auto& xLocal) mutable -> auto {
            // return a linear interpolation of the vertex shift values
            std::vector<GlobalCoordinate> shift( dim+1 );
            for ( std::size_t i = 0; i < entity.subEntities( dim ); ++i )
            {
                const auto& vertex = entity.template subEntity<dim>( i );
                shift[i] = shifts[ iindexSet.index( vertex ) ];
            }
            const Dune::MultiLinearGeometry<double, dim, dimw> interpolation(entity.type(), shift);
            return interpolation.global(xLocal);
        };
        return ret;
    }
    """
    from dune.fem.function import cppFunction
    cppFunc = cppFunction(igrid, name="interfaceEdgeMovement", order=1, fctName="interfaceEdgeMovement", includes=io.StringIO(code), args=[igrid, shifts])
    from dune.fem.space import lagrange
    space = lagrange(igrid, dimRange=igrid.dimWorld, order=1)
    return space.interpolate(cppFunc, name="interfaceEdgeMovement")



import sys

if len(sys.argv) > 1:
    name = sys.argv[1]
else:
    name = 'grid.msh'

tikz = []
readNodes = False
readElements = False
edges = set()

file = open(name, 'r')
for line in file.readlines():
    line = line.strip()
    if line == '$Nodes':
        readNodes = True
    if line == '$EndNodes':
        readNodes = False
    if line == '$Elements':
        readElements = True
    if line == '$EndElements':
        readElements = False

    if readNodes:
        data = line.split(' ')
        if len(data) != 4:
            continue
        i, x, y, z = data
        tikz += ['\coordinate ('+i+') at ('+x+','+y+');']

    if readElements:
        data = line.split(' ')
        # normal edges
        if len(data) == 8:
            _, _, _, _, _, a, b, c = line.split(' ')
            for edge in [(a, b), (a, c), (b, c)]:
                p, q = sorted(edge)
                if not (p, q) in edges:
                    tikz += ['\draw[very thin] ('+p+') -- ('+q+');']
                    edges.add((p, q));

        # interface edges
        if len(data) == 7:
            _, _, _, o, p, a, b = line.split(' ')
            if int(o) >= 10 or int(p) >= 10:
                tikz += ['\draw[very thick] ('+a+') -- ('+b+');']

out = open(name+'.tikz', 'w')
out.write("\n".join(tikz))
out.close()



import sys
import subprocess
from argparse import ArgumentParser

def main(arguments=None):
    parser = ArgumentParser(description='Run dune-mmesh commands', prog='dune-mmesh')
    subparsers = parser.add_subparsers(dest='command')

    parserConfigure = subparsers.add_parser('test', help='Test dune-mmesh installation')

    ret = 0
    args = parser.parse_args(arguments)
    if args.command == 'test':
        subprocess.run([sys.executable, '-m', 'dune.mmesh.test'], stdout=sys.stdout)

    sys.exit(ret)


if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python
# coding: utf-8

# # Moving and adapting
#
# This is an example of how to move the interface and adapt the mesh.

# We implement the finite volume moving mesh method presented in [CMR+18]_. 
# 
# .. [CMR+18] C. Chalons, J. Magiera, C. Rohde, M. Wiebe. A Finite-Volume Tracking Scheme for Two-Phase Compressible Flow. Theory, Numerics and Applications of Hyperbolic Problems I, pp. 309--322, 2018.

# ## Grid creation
# 
# We use the [vertical](grids/vertical.rst) grid file that contains an interface $\Gamma = {0.5} \times [0, 1]$ embedded in a domain $\Omega = [0,1]^2$.
# For this example, we have to construct an _adaptive_ leaf grid view and we will need to obtain the hierarchical grid object.

# In[1]:


from dune.grid import reader
from dune.mmesh import mmesh
from dune.fem.view import adaptiveLeafGridView as adaptive
dim = 2
file = "grids/vertical.msh"
gridView = adaptive( mmesh((reader.gmsh, file), dim) )
hgrid = gridView.hierarchicalGrid
igridView = hgrid.interfaceGrid


# ## Problem
# 
# Let us consider the following transport problem.
# \begin{align}
# u_t + \operatorname{div} f(u) = 0, & \qquad \text{in } \Omega \times [0,T], \\
# u(\cdot, 0) = u_0, & \qquad \text{in } \Omega
# \end{align}
# where
# \begin{align}
# f(u) &= [1,0]^T u, \\
# u_0(x,y) &= (0.5+x) \chi_{x<0.5}.
# \end{align}
# 
# Further, the interface is supposed to move with the transport speed in $f$, i.e. $m = [1,0]^T$.
# \begin{align*}
# \renewcommand{\jump}[1]{[\mskip-5mu[ #1 ]\mskip-5mu]}
# \end{align*}

# In[2]:


from ufl import *
from dune.ufl import Constant

tEnd = 0.4

def speed():
    return as_vector([1.0, 0.0])

def movement(x):
    return as_vector([1.0, 0.0])

def f(u):
    return speed() * u

def u0(x):
    return conditional(x[0] < 0.5, 0.5+x[0], 0.0)

def uexact(x, t):
    return u0( x - t * speed() )


# ## Finite Volume Moving Mesh Method
# 
# We use a Finite Volume Moving Mesh method to keep the discontinuity sharp. It can be formulated by
# \begin{align}
# \int_\Omega (u^{n+1} |det(\Psi)| - u^n) v\ dx + \Delta t \int_\mathcal{F} \big( g(u^{n}, n) - h(u^{n}, n) \big) \jump{v}\ dS = 0
# \end{align}
# where $\Psi := x + \Delta t s$ and s is a linear interpolation of the interface’s vertex movement m on the bulk triangulation.
# 
# The numerical fluxes $g(u, n)$ and $h(u, n)$ are assumed to be consistent with the flux functions $f(u) \cdot n$ and $u s \cdot n$, respectively.

# In[3]:


from dune.fem.space import finiteVolume

space = finiteVolume(gridView)
u = TrialFunction(space)
v = TestFunction(space)

x = SpatialCoordinate(space)
n = FacetNormal(space)

uh = space.interpolate(u0(x), name="uh")
uh_old = uh.copy()


# In[4]:


import numpy as np
from dune.geometry import vertex
from dune.mmesh import edgeMovement

def getShifts():
    mapper = igridView.mapper({vertex: 1})
    shifts = np.zeros((mapper.size, dim))
    for v in igridView.vertices:
        shifts[ mapper.index(v) ] = as_vector(movement( v.geometry.center ))
    return shifts

em = edgeMovement(gridView, getShifts())
t = Constant(0, name="time")

def g(u, n):
    sgn = inner(speed(), n('+'))
    return inner( conditional( sgn > 0, f( u('+') ), f( u('-') ) ), n('+') )

def gBnd(u, n):
    sgn = inner(speed(), n)
    return inner( conditional( sgn > 0, f(u), f(uexact(x, t)) ), n )

def h(u, n):
    sgn = inner(em('+'), n('+'))
    return conditional( sgn > 0, sgn * u('+'), sgn * u('-') )


# In[5]:


from dune.fem.scheme import galerkin

tau = Constant(0, name="tau")
detPsi = abs(det(nabla_grad(x + tau * em)))

a = (u * detPsi - uh_old) * v * dx
a += tau * (g(uh_old, n) - h(uh_old, n)) * jump(v) * dS
a += tau * gBnd(uh_old, n) * v * ds

scheme = galerkin([a == 0], solver=("suitesparse","umfpack"))


# Compute CFL time step using cell edge lengths
l, ll = TrialFunction(space), TestFunction(space)
b = (l - MinCellEdgeLength(space)) * ll * dx
lscheme = galerkin([b == 0], solver=("suitesparse","umfpack"))
lh = space.interpolate(0, name="lh")

def dtCFL():
    lscheme.solve(lh)
    lMin = np.min(lh.as_numpy)
    return 0.5 * lMin / speed()[0]


# ## Timeloop without adaptation
# 
# For comparison, we run the finite volume scheme once without adaptation of the mesh.

# In[6]:


from time import time
from dune.fem.function import integrate
from dune.fem.plotting import plotPointData as plot
import matplotlib.pyplot as plt
plots = 4
fig, axs = plt.subplots(1, plots, figsize=(3*plots,3))
def shouldPlot(a):
    t_a = a * tEnd / (plots-1)
    return t.value > t_a and t.value - t_a <= tau.value

runtime = 0

uh.interpolate(u0(x))

def L2Error(uh):
    u = uexact(x, t)
    return sqrt(integrate(gridView, dot(uh-u, uh-u), order=5))

l2s = []

# Disable edge movement
em.interpolate([0,0])

tau.value = dtCFL()
print(f"dt = {tau.value:.4f}")

i = 0
t.assign(0)
while t.value < tEnd:
    runtime -= time()

    t.value += tau.value

    uh_old.assign(uh)
    scheme.solve(target=uh)
    l2s += [(t.value, L2Error(uh))]

    runtime += time()

    i += 1
    for a in range(plots):
        if shouldPlot(a):
            plot(uh, figure=(fig, axs[a]), clim=[0,1], colorbar=None)

print(f"Runtime: {runtime:.3f}s")

L2 = L2Error(uh)
print(f"L2-Error: {L2:.3f}")


# ## Timeloop with adaptation
# 
# We need to set the `"fem.adaptation.method"` parameter to `"callback"` in order to use the non-hierarchical adaptation strategy of Dune-MMesh. Then, within the time loop, we can adapt the mesh according to the following strategy.

# In[7]:


from dune.fem import parameter, adapt
parameter.append( { "fem.adaptation.method": "callback" } )
fig.clear()
fig, axs = plt.subplots(1, plots, figsize=(3*plots,3))
runtimeAdapted = 0
dtmin = 1e3

uh.interpolate(u0(x))

l2sAdapted = []

i = 0
t.assign(0)
while t.value < tEnd:
    runtimeAdapted -= time()

    tau.value = dtCFL()
    dtmin = min(dtmin, tau.value)

    t.value += tau.value
    
    mark = hgrid.markElements()
    ensure = hgrid.ensureInterfaceMovement(getShifts()*tau.value)
    if mark or ensure:
        adapt([uh])
    
    shifts = getShifts()
    em.assign(edgeMovement(gridView, shifts))

    uh_old.assign(uh)
    scheme.solve(target=uh)

    hgrid.moveInterface(shifts*tau.value)
    l2sAdapted += [(t.value, L2Error(uh))]

    runtimeAdapted += time()

    i += 1
    for a in range(plots):
        if shouldPlot(a):
            plot(uh, figure=(fig, axs[a]), clim=[0,1], colorbar=None)


print(f"dt_min = {dtmin:.4f}")
print(f"Runtime: {runtimeAdapted:.3f}s")

L2Adapted = L2Error(uh)
print(f"L2-Error: {L2Adapted:.3f}")

print(f"\nRuntime factor: {runtimeAdapted / runtime:.2f}x")
print(f"Error improvement: {L2 / L2Adapted:.2f}x")

import matplotlib.pyplot as plt
fig = plt.figure()
plt.plot([l2[0] for l2 in l2s], [l2[1] for l2 in l2s], label="Without adaptation")
plt.plot([l2[0] for l2 in l2sAdapted], [l2[1] for l2 in l2sAdapted], label="Adapted")
plt.legend(loc='best')
plt.title('L2-Error')
plt.show()



#!/usr/bin/env python
# coding: utf-8

# # Mixed-dimensional problem

# A strength of Dune-MMesh is its application to mixed-dimensional problems.
# So, let us consider a mixed-dimensional Burger's equation on a domain with a lower-dimensional interface $\Gamma$ that represents a physical domain with aperture $d > 0$.
#
# Find $u, u_\Gamma$ s.t
# \begin{align}
# \renewcommand{\jump}[1]{[\mskip-5mu[ #1 ]\mskip-5mu]}
# u_t + \operatorname{div} \mathbf{F}(u) &= 0, &&\quad \text{in } \Omega, \\
# (d u_\Gamma)_t + \operatorname{div} c \mathbf{F}(u_\Gamma) &= \jump{\mathbf{F}(u) \cdot \mathbf{n}}, &&\quad \text{in } \Gamma,
# \end{align}
# where $\mathbf{F}(u) = \frac{1}{2} u^2 \mathbf{a}$ with $\mathbf{a} = [1,0]^T$ and $c > 0$ is a speed scaling for the interface problem.
#
# We impose the interior boundary condition for the bulk problem
# \begin{align}
# F(u) &= F(u_\Gamma), \quad \text{on } \Gamma, \\
# \end{align}
# and consider inflow from left on the [T-junction](grids/tjunction.rst) grid.

# In[1]:


from dune.grid import reader
from dune.mmesh import mmesh

file = "grids/tjunction.msh"

gridView = mmesh((reader.gmsh, file), 2)
igridView = gridView.hierarchicalGrid.interfaceGrid

# In[2]:


from ufl import *
from dune.ufl import Constant

d = Constant(0.01, name="d")
a = as_vector([1, 0])
c = Constant(100, name="c")

def F(u):
    return 0.5 * u**2 * a

def upwind(u_l, u_r, n):
    return inner( conditional( inner(a, n) > 0, F(u_l), F(u_r) ), n )

dt = 0.25
tau = Constant(dt, name="tau")

# ## Bulk problem
#
# We use a finite volume space and define the UFL form of the bulk problem. Hereby, we used the interface indicator `I` to distinguish between interface and non-interface facets.

# In[3]:


from dune.fem.space import finiteVolume
space = finiteVolume(gridView)

u  = TrialFunction(space)
uu = TestFunction(space)

x = SpatialCoordinate(space)
n = FacetNormal(space)

left = conditional(x[0] < 1e-6, 1, 0)
uD = 1

uh = space.interpolate(0, name="uh")
uhOld = uh.copy()

from dune.mmesh import interfaceIndicator
I = interfaceIndicator(igridView)

A  = (u - uhOld) * uu * dx
A += tau * upwind(u('+'), u('-'), n('+')) * jump(uu) * (1-I)*dS
A += tau * upwind(u, uD, n) * uu * left * ds

# ## Interface problem
#
# Similarly, we define the UFL form of the interface problem.

# In[4]:


ispace = finiteVolume(igridView)

u_g  = TrialFunction(ispace)
uu_g = TestFunction(ispace)

n_g = FacetNormal(ispace)

uh_g = ispace.interpolate(0, name="uh_g")
uhOld_g = uh_g.copy()

A_g  = d * (u_g - uhOld_g) * uu_g * dx
A_g += tau * c * upwind(u_g('+'), u_g('-'), n_g('+')) * jump(uu_g) * dS

# ## Coupling condition
#
# The coupling condition can be incorporated using the trace and skeleton functionality.

# In[5]:


from dune.mmesh import skeleton, trace, normals

u_gamma = avg(skeleton(uh_g))

A += tau * upwind(u('+'), u_gamma, n('+')) * uu('+') * I*dS
A += tau * upwind(u('-'), u_gamma, n('-')) * uu('-') * I*dS


traceu = trace(uh)
inormal = normals(igridView)

A_g -= tau * upwind(traceu('+'), u_g, -inormal) * uu_g * dx
A_g -= tau * upwind(traceu('-'), u_g, -inormal) * uu_g * dx

# ## Monolithic solution strategy
#
# Dune-MMesh provides coupled solution strategies to solve bulk and interface schemes together, either iteratively or monolithically. Here, we choose the monolithicSolve method that implements a mixed-dimensional Newton method.

# In[6]:


from dune.fem.scheme import galerkin
scheme   = galerkin([A   == 0])
scheme_g = galerkin([A_g == 0])

from dune.mmesh import monolithicSolve
def solve():
    uhOld.assign(uh)
    uhOld_g.assign(uh_g)
    monolithicSolve(schemes=(scheme, scheme_g), targets=(uh, uh_g), verbose=False)

# In[7]:


import matplotlib.pyplot as plt
from dune.fem.plotting import plotPointData as plot

fig, axs = plt.subplots(1, 4, figsize=(20,5))

for i in range(4):
    solve()

    plot(uh, figure=(fig, axs[i]), gridLines=None, clim=[0,1])
    plot(uh_g, figure=(fig, axs[i]), linewidth=0.04, clim=[0,1], colorbar=None)
